// $Id: CvsInterface.cpp,v 1.94 2012/09/22 19:27:30 hpereira Exp $

/******************************************************************************
*
* Copyright (C) 2002 Hugo PEREIRA <mailto: hugo.pereira@free.fr>
*
* This is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
*
* This software is distributed in the hope that it will be useful, but WITHOUT
* Any WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License along with
* software; if not, write to the Free Software Foundation, Inc., 59 Temple
* Place, Suite 330, Boston, MA  02111-1307 USA
*
*
*******************************************************************************/

#include "CvsInterface.h"

#include "Command.h"
#include "CvsLockDialog.h"
#include "Debug.h"
#include "InformationDialog.h"
#include "IOString.h"
#include "Singleton.h"
#include "TextFormat.h"
#include "XmlOptions.h"

#include <QtGui/QApplication>
#include <QtCore/QRegExp>
#include <QtCore/QTextStream>

#include <algorithm>

//___________________________________________________________________
CvsInterface::CvsInterface( QObject *parent ):
    QObject( parent ),
    Counter( "CvsInterface" ),
    process_( 0 ),
    ignoreLocks_( false )
{ Debug::Throw( "CvsInterface::CvsInterface.\n" ); }

//___________________________________________________________________
void CvsInterface::logRevisions( const File& file )
{

    Debug::Throw( "CvsInterface::logRevisions.\n" );
    assert( !file.isDirectory() );

    _addCommand( CvsCommand( CvsCommand::LOG ) << file );
    return;

}

//_______________________________________________________________________
void CvsInterface::logStatus( const File& path, const FileInfo::List& files )
{

    Debug::Throw() << "CvsInterface::logStatus - path: " << path << endl;

    CvsCommand command = CvsCommand( CvsCommand::STATUS ) << path;
    for( FileInfo::List::const_iterator iter = files.begin(); iter != files.end(); ++iter )
    { command << iter->file(); }

    _addCommand( command );

    return;

}

//_______________________________________________________________________
void CvsInterface::fakeUpdate( const File& path, const FileInfo::List& files )
{

    CvsCommand command = CvsCommand( CvsCommand::FAKE_UPDATE ) << path;
    for( FileInfo::List::const_iterator iter = files.begin(); iter != files.end(); ++iter )
    { command << iter->file(); }
    _addCommand( command );

    return;

}

//_______________________________________________________________________
void CvsInterface::checkOutRevision( const File& file, const QString& revision )
{

    Debug::Throw() <<  "CvsInterface::checkOutRevision - file: " << file << " revision: " << revision << endl;
    assert( !file.isDirectory() );

    _addCommand( CvsCommand( CvsCommand::CHECKOUT_REVISION ) << file << revision );
    return;
}

//_______________________________________________________________________
void CvsInterface::checkOut( const File& path, const QString& cvs_root, const QStringList& modules, const File& local_path )
{
    Debug::Throw( "CvsInterface::checkOut.\n" );

    if( modules.empty() ) return;
    else if( modules.size() == 1 )
    {
        _addCommand( CvsCommand( CvsCommand::CHECKOUT )
            << path
            << cvs_root
            << modules.front()
            << local_path );
    } else for( QStringList::const_iterator iter = modules.begin(); iter != modules.end(); ++iter ) {

        // if several modules are provided, the local_path is ignored and modules are checked-out as-is
        _addCommand(
            CvsCommand( CvsCommand::CHECKOUT )
            << path << cvs_root << *iter << "" );

    }

    return;
}

//_______________________________________________________________________
void CvsInterface::update( const File& path, const FileInfo::List& files )
{

    Debug::Throw( "CvsInterface::update.\n" );
    CvsCommand command = CvsCommand( CvsCommand::UPDATE ) << path;
    for( FileInfo::List::const_iterator iter = files.begin(); iter != files.end(); ++iter )
    { command << iter->file(); }
    _addCommand( command );
    return;
}

//_______________________________________________________________________
void CvsInterface::add( const File& path, const FileInfo::List& files )
{

    Debug::Throw( "CvsInterface::add.\n" );
    CvsCommand command = CvsCommand( CvsCommand::ADD ) << path;
    for( FileInfo::List::const_iterator iter = files.begin(); iter != files.end(); ++iter )
    { command << iter->file(); }
    _addCommand( command );
    return;
}

//_______________________________________________________________________
void CvsInterface::remove( const File& path, const FileInfo::List& files )
{

    Debug::Throw( "CvsInterface::remove.\n" );
    CvsCommand command = CvsCommand( CvsCommand::REMOVE ) << path;
    for( FileInfo::List::const_iterator iter = files.begin(); iter != files.end(); ++iter )
    { command << iter->file(); }
    _addCommand( command );
    return;
}

//_______________________________________________________________________
void CvsInterface::commit( const File& path, const File& comments, const FileInfo::List& files )
{

    Debug::Throw( "CvsInterface::commit.\n" );
    CvsCommand command = CvsCommand( CvsCommand::COMMIT ) << path << comments;
    for( FileInfo::List::const_iterator iter = files.begin(); iter != files.end(); ++iter )
    { command << iter->file(); }
    _addCommand( command );

    return;
}

//____________________________________________________________
bool CvsInterface::isCvsPath( File path ) const
{
    Debug::Throw( "CvsInterface::isCvsPath" );
    if( !( path.exists() && path.isDirectory() ) ) return false;
    return
        File( "CVS/Root" ).addPath( path ).exists() &&
        File( "CVS/Repository" ).addPath( path ).exists();
}

//____________________________________________________________
bool CvsInterface::isCvsFile( File file ) const
{
    Debug::Throw( "CvsInterface::isCvsFile" );
    return std::find( entries_.begin(), entries_.end(), file ) != entries_.end();
}

//____________________________________________________________
bool CvsInterface::updateRepositoryPath( File path )
{
    Debug::Throw( "CvsInterface::updateRepositoryPath.\n" );

    // try load root directory
    {
        QFile in( File( "CVS/Root" ).addPath( path ) );
        if( in.open( QIODevice::ReadOnly ) )
        {
            QString buffer( in.readAll() );
            QTextStream in_stream( &buffer );
            in_stream >> root_;
            if( in_stream.status() != QTextStream::Ok ) return false;
        } else return false;

        QString buffer;
        QTextStream( &buffer ) << "[root: " << root_ << "]" << endl;
        emit messageAvailable( buffer );
    }

    // try load repository
    {
        QFile in( File( "CVS/Repository" ).addPath( path ) );
        if( in.open( QIODevice::ReadOnly ) )
        {
            QString buffer( in.readAll() );
            QTextStream in_stream( &buffer );
            in_stream >> repository_;
            if( in_stream.status() != QTextStream::Ok ) return false;
        } else return false;

        QString buffer;
        QTextStream( &buffer ) << "[repository: " << repository_ << "]" << endl;
        emit messageAvailable( buffer );
    }

    return true;

}

//____________________________________________________________
bool CvsInterface::updateEntries( File path )
{
    Debug::Throw( "CvsInterface::updateEntries.\n" );

    // try load entries
    {

        entries_.clear();
        QFile in( File( "CVS/Entries" ).addPath( path ) );
        if( !in.open( QIODevice::ReadOnly ) ) return false;
        QStringList lines( QString( in.readAll() ).split( "\n" ) );
        for( QStringList::iterator iter = lines.begin(); iter != lines.end(); ++iter )
        {

            QStringList string_list( iter->split( "/" ) );
            if( string_list.size() < 2 ) continue;

            entries_.push_back( File( string_list[1] ).addPath( path ) );

        }

    }

    return true;

}

//____________________________________________________________
void CvsInterface::abort( int type )
{
    Debug::Throw( "CvsInterface::abort.\n" );
    QList<CvsCommand>::iterator iter;

    if( !commands_.size() ) return;

    // loop over commands, keep only the ones that do not match the requested type
    QList<CvsCommand> new_commands;
    for( QList<CvsCommand>::iterator iter = commands_.begin(); iter != commands_.end(); ++iter )
    {
        if( !(iter->type() & type ) ) new_commands.push_back( *iter );
        else if( iter != commands_.begin() || _process().state() == QProcess::NotRunning ) emit commandRemoved( *iter );
    }

    // if process is running and the last command should be removed,
    // stop the process, but add the command to the list, since it is needed
    // after the process is stopped
    if( _process().state() != QProcess::NotRunning && ( commands_.front().type() & type ) )
    {
        new_commands.push_front( commands_.front() );
        commands_ = new_commands;
        QString buffer;
        QTextStream( &buffer ) << "[" << commands_.front().name() << " aborted]" << endl << endl;
        emit messageAvailable( buffer );
        _process().kill();
    } else commands_ = new_commands;

    return;
}

//____________________________________________________________
void CvsInterface::abort( CvsCommand command )
{

    Debug::Throw() << "CvsInterface::abort - command id: " << command.id() << " type: " << command.name() << endl;
    QList<CvsCommand>::iterator iter( std::find( commands_.begin(), commands_.end(), command ) );
    if( iter == commands_.end() ) return;

    if( command == commands_.front() && _process().state() != QProcess::NotRunning )
    {

        QString buffer;
        QTextStream( &buffer ) << "[" << command.name() << " aborted]" << endl << endl;
        emit messageAvailable( buffer );
        _process().kill();

    } else {
        emit commandRemoved( *iter );
        commands_.erase( iter );
    }

}

//____________________________________________________________
void CvsInterface::_error( QProcess::ProcessError error )
{

    Debug::Throw( "CvsInterface::_error.\n" );
    if( commands_.empty() ) return;

    QString buffer;
    QTextStream( &buffer ) << commands_.front().name() << ": " << CustomProcess::errorMessage( error );
    Debug::Throw() << buffer << endl;

    // it is needed to complete the command manually when
    // error message is "failed to start" otherwise the busy widget
    // will stay busy forever. Also, display a popup
    if( error == QProcess::FailedToStart )
    {
        InformationDialog( qApp->activeWindow(),  buffer ).exec();
        _commandCompleted( 0, QProcess::NormalExit );
    }

    return;
}

//___________________________________________________________________
void CvsInterface::_readStdout(  void )
{

    _process().setReadChannel( QProcess::StandardOutput );
    stdoutBuffer_.append( IOString( _process() ) );
    _parseStdoutBuffer();

}

//___________________________________________________________________
void CvsInterface::_readStderr(  void )
{

    _process().setReadChannel( QProcess::StandardError );
    stderrBuffer_.append( IOString( _process() ) );
    _parseStderrBuffer();

}

//___________________________________________________________________
void CvsInterface::_parseStdoutBuffer( void )
{

    Debug::Throw( "CvsInterface::_parseStdoutBuffer.\n" );
    if( commands_.empty() ) return;

    // retrieve command and redirect
    // only STATUS and FAKE_UPDATE runs on fly.
    // other commands wait for the process to be completed.
    switch( commands_.front().type() )
    {

        case CvsCommand::LOG:
        _logCommandAvailable();
        break;

        case CvsCommand::STATUS:
        _statusCommandAvailable();
        break;

        case CvsCommand::FAKE_UPDATE:
        _fakeUpdateCommandAvailable();
        break;

        case CvsCommand::CHECKOUT:
        _checkOutCommandAvailable();
        break;

        case CvsCommand::UPDATE:
        _updateCommandAvailable();
        break;

        case CvsCommand::COMMIT:
        _commitCommandAvailable();
        break;

        default:
        break;
    }

}


//___________________________________________________________________
void CvsInterface::_parseStderrBuffer( void )
{

    Debug::Throw( "CvsInterface::_parseStdoutBuffer.\n" );
    if( commands_.empty() ) return;

    _checkLock( stderrBuffer_.text().mid( stderrBuffer_.position(), stderrBuffer_.text().length() ) );

    // retrieve command and redirect
    // only STATUS and FAKE_UPDATE runs on fly.
    // other commands wait for the process to be completed.
    switch( commands_.front().type() )
    {

        case CvsCommand::COMMIT:
        _commitErrorAvailable();
        break;

        default:
        break;
    }

}

//___________________________________________________________________
void CvsInterface::_commandCompleted(  int code, QProcess::ExitStatus status )
{

    Debug::Throw() << "CvsInterface::_commandCompleted - code: " << code << " status: " << status << endl;

    if( commands_.empty() ) return;
    if( status == QProcess::NormalExit )
    {

        // retrieve command and redirect
        switch( commands_.front().type() )
        {

            case CvsCommand::CHECKOUT_REVISION:
            _revisionCheckOutCommandCompleted();
            break;

            case CvsCommand::COMMIT:
            _commitCommandCompleted();
            break;

            default:
            break;
        }

    }

    if( status == QProcess::NormalExit )
    {
        QString buffer;
        QTextStream( &buffer ) << "[" << commands_.front().name() << " completed]" << endl << endl;
        emit messageAvailable( buffer );
    }

    //! emit signal
    emit commandCompleted( commands_.front() );

    // remove command from list
    // run next command if any
    commands_.pop_front();
    if( !commands_.empty() ) _run();

    return;

}

//____________________________________________________________
void CvsInterface::_resetBuffers( void )
{
    Debug::Throw( "CvsInterface::_resetBuffers.\n" );
    stdoutBuffer_.clear();
    stderrBuffer_.clear();
}

//_______________________________________________________________________
void CvsInterface::_addCommand( CvsCommand command )
{

    // add to list, emit signal, and run.
    commands_.push_back( command );
    emit commandAdded( command );

    if( commands_.size() == 1 ) _run();
}

//_______________________________________________________________________
void CvsInterface::_run( void )
{
    Debug::Throw( "CvsInterface::_run.\n" );
    if( commands_.empty() ) return;

    // reset buffer
    _resetBuffers();
    _resetProcess();

    // reset ignoreLocks_ flag
    ignoreLocks_ = false;

    // retrieve pending command
    const CvsCommand& command( commands_.front() );
    emit commandStarted( command );

    // retrieve cvs command from options
    Command cvs( XmlOptions::get().raw( "CVS" ) );

    // start process with matching command type.
    switch (command.type())
    {

        case CvsCommand::LOG:
        {
            QString buffer;
            QTextStream(&buffer) << "[" << command.name() << " " <<  command.argument(0) << "]" << endl;
            emit messageAvailable( buffer );

            _newProcess().setWorkingDirectory( QFileInfo( command.argument(0) ).absolutePath() );
            _process().start( cvs << "log" << QFileInfo( command.argument(0) ).fileName() );

        }
        break;

        case CvsCommand::STATUS:
        { _runStatusCommand(); }
        break;

        case CvsCommand::FAKE_UPDATE:
        { _runFakeUpdateCommand(); }
        break;

        case CvsCommand::CHECKOUT_REVISION:
        {

            QString buffer;
            QTextStream(&buffer) << "[" << command.name() << " " << command.argument(0) << "]" << endl;
            emit messageAvailable( buffer );

            _newProcess().setWorkingDirectory( QFileInfo( command.argument(0) ).absolutePath() );
            _process().start(
                cvs << "update"
                << "-r" << command.argument(1)
                << "-p" << QFileInfo( command.argument(0) ).fileName() );
        }
        break;

        case CvsCommand::CHECKOUT:
        { _runCheckOutCommand(); }
        break;

        case CvsCommand::UPDATE:
        { _runUpdateCommand(); }
        break;

        case CvsCommand::ADD:
        {
            QString buffer;
            QTextStream( &buffer ) << "[" << command.name() << " " << command.argument(0) << "]" << endl;
            emit messageAvailable( buffer );

            QString path( command.argument(0) );
            assert( QFileInfo( path ).isDir() );
            _newProcess().setWorkingDirectory( path );

            QStringList command_string( cvs << "add" );
            for( int i=1; i<command.argumentCount(); i++ )
            { command_string << QFileInfo( command.argument(i) ).fileName(); }

            _process().start( command_string );

        }
        break;

        case CvsCommand::REMOVE:
        {
            QString buffer;
            QTextStream( &buffer ) << "[" << command.name() << " " << command.argument(0) << "]" << endl;
            emit messageAvailable( buffer );

            QString path( command.argument(0) );
            assert( QFileInfo( path ).isDir() );
            _newProcess().setWorkingDirectory( path );

            QStringList command_string( cvs << "remove" );
            for( int i=1; i<command.argumentCount(); i++ )
            { command_string << QFileInfo(command.argument(i)).fileName(); }

            _process().start( command_string );

        }
        break;

        case CvsCommand::COMMIT:
        { _runCommitCommand(); }
        break;

        default: break;

    }

    return;

}

//_______________________________________________________________________
void CvsInterface::_runStatusCommand( void )
{
    // retrieve pending command
    CvsCommand& command( commands_.front() );

    // retrieve path for update command
    QString path( command.argument(0) );
    assert( QFileInfo(path).isDir() );

    QString buffer;
    QTextStream(&buffer) << "[" << command.name() << " " <<  command.argument(0) << "]" << endl;
    emit messageAvailable( buffer );

    // do nothing if path is not CVS
    if( !isCvsPath( File( path ) ) )  _commandCompleted( 0, QProcess::NormalExit );
    else
    {
        // retrieve cvs command
        Command cvs( XmlOptions::get().raw( "CVS" ) );
        _newProcess().setWorkingDirectory( path );
        QStringList command_string( cvs << "status" << "-l" );
        for( int i=1; i<command.argumentCount(); i++ )
        {

            // no need to run on subdirectories, since they will not
            // appear in the file browser anyway
            QFileInfo file( command.argument(i) );
            if( !file.isDir() )
            { command_string << file.fileName(); }

        }

        _process().start( command_string );
    }

}

//_______________________________________________________________________
void CvsInterface::_runFakeUpdateCommand( void )
{
    // retrieve pending command
    CvsCommand& command( commands_.front() );

    // retrieve path for update command
    QString path( command.argument(0) );
    assert( QFileInfo( path ).isDir() );

    QString buffer;
    QTextStream(&buffer) << "[" << command.name() << " " <<  command.argument(0) << "]" << endl;
    emit messageAvailable( buffer );

    // do nothing if path is not CVS
    if( !isCvsPath( path ) )  _commandCompleted( 0, QProcess::NormalExit );
    else
    {
        // retrieve cvs command
        Command cvs( XmlOptions::get().raw( "CVS" ) );
        _newProcess().setWorkingDirectory( path );
        QStringList command_string( cvs << "-n" << "update" << "-l" );
        for( int i=1; i<command.argumentCount(); i++ )
        {
            // no need to run on subdirectories, since they will not
            // appear in the file browser anyway
            QFileInfo file( command.argument(i) );
            if( !file.isDir() )
            { command_string << file.fileName(); }
        }
        _process().start( command_string );

    }

    return;
}

//_______________________________________________________________________
void CvsInterface::_runCheckOutCommand( void )
{
    Debug::Throw( "CvsInterface::_runCheckOutCommand.\n" );

    // retrieve pending command
    CvsCommand& command( commands_.front() );
    Command cvs( XmlOptions::get().raw( "CVS" ) );

    // retrieve path for update command
    QString path( command.argument(0) );
    assert( QFileInfo( path ).isDir() );

    QString cvs_root( command.argument(1) );
    QString module( command.argument(2) );
    QString local_path( command.argument(3) );

    QString buffer;
    QTextStream(&buffer) << "[" << command.name() << " " << module << "]" << endl;
    emit messageAvailable( buffer );

    QStringList command_string( cvs );
    if( !cvs_root.isEmpty() ) command_string << "-d" << cvs_root;
    command_string << "checkout";

    if( !local_path.isEmpty() ) command_string << "-d" << local_path;
    command_string << module;

    _newProcess().setWorkingDirectory( path );
    _process().start( command_string );

    return;
}


//_______________________________________________________________________
void CvsInterface::_runUpdateCommand( void )
{
    Debug::Throw( "CvsInterface::_runUpdateCommand.\n" );

    // retrieve pending command
    CvsCommand& command( commands_.front() );
    Command cvs( XmlOptions::get().raw( "CVS" ) );

    // retrieve path for update command
    QString path( command.argument(0) );
    assert( QFileInfo( path ).isDir() );

    QString buffer;
    QTextStream(&buffer) << "[" << command.name() << " " << path << "]" << endl;
    emit messageAvailable( buffer );

    // directories that are found in the arguments
    // must be handled separately by CVS.
    // they are kept aside and processed via dedicated commands
    // added (recursively) from here .
    QStringList subdirectories;
    QStringList files;
    for( int i=1; i<command.argumentCount(); i++ )
    {
        QString file( command.argument(i) );
        if( QFileInfo( file ).isDir() )
        {
            Debug::Throw() << "CvsInterface::_runUpdateCommand - subdirectory: " << file << endl;
            subdirectories.push_back( file );
        } else files.push_back( file );
    }

    // run current command if any files are in the arguments
    // otherwise set command as completed
    if( !files.empty() || subdirectories.empty() )
    {

        // prepare command
        QStringList command_string( cvs << "update" );

        // add files
        for( QStringList::iterator iter = files.begin(); iter != files.end(); ++iter )
        { command_string << QFileInfo( *iter ).fileName(); }

        // run process
        _newProcess().setWorkingDirectory( path );
        _process().start( command_string );

    } else _commandCompleted( 0, QProcess::NormalExit );

    // add dedicated commands for found subdirectories
    if( !subdirectories.empty() )
    {
        for( QStringList::iterator iter = subdirectories.begin(); iter != subdirectories.end(); ++iter )
        {
            // need to make sure the trailing "/" is needed or not.
            _addCommand( CvsCommand( CvsCommand::UPDATE ) << (*iter + "/") );
        }
    }

    return;

}

//______________________________________________________________
void CvsInterface::_runCommitCommand( void )
{
    Debug::Throw( "CvsInterface::_runCommitCommand.\n" );

    // retrieve pending command
    CvsCommand& command( commands_.front() );
    Command cvs( XmlOptions::get().raw( "CVS" ) );

    // retrieve path for update command
    QString path = command.argument(0);
    assert( QFileInfo( path ).isDir() );

    // try retrieve comments file
    QString comments( command.argument(1) );
    QString buffer;
    QTextStream(&buffer) << "[" << command.name() << " " << path << "]" << endl;
    emit messageAvailable( buffer );

    // directories that are found in the arguments
    // must be handled separately by CVS.
    // they are kept aside and processed via dedicated commands
    // added (recursively) from here .
    QStringList subdirectories;
    QStringList files;
    for( int i=2; i<command.argumentCount(); i++ )
    {
        QString file( command.argument(i) );
        if( QFileInfo( file ).isDir() )
        {

            Debug::Throw() << "CvsInterface::_runCommitCommand - subdirectory: " << file << endl;
            subdirectories.push_back( file );

        } else files.push_back( file );

    }

    // run current command if any files are in the arguments
    // otherwise set command as completed
    if( !files.empty() || subdirectories.empty() )
    {

        // prepare command
        QStringList command_string( cvs << "commit" );

        // add comments file, if any
        if( comments.isEmpty() || !QFileInfo( comments ).exists() ) command_string << "-m" << "\"no message\"";
        else command_string << "-F" << comments;

        // add files
        for( QStringList::iterator iter = files.begin(); iter != files.end(); ++iter )
        { command_string << QFileInfo( *iter ).fileName(); }

        // run process
        _newProcess().setWorkingDirectory( path );
        _process().start( command_string );

    } else _commandCompleted( 0, QProcess::NormalExit );

    // add dedicated commands for found subdirectories
    if( !subdirectories.empty() )
    {
        for( QStringList::iterator iter = subdirectories.begin(); iter != subdirectories.end(); ++iter )
        {
            // need to make sure the trailing slash is needed
            _addCommand( CvsCommand( CvsCommand::COMMIT ) << (*iter + "/") << comments );
        }
    }

}

//_______________________________________________________________________
bool CvsInterface::_checkLock( const QString& buffer )
{
    Debug::Throw( "CvsInterface::_checkLock.\n" );

    static const QString user_start_tag = "waiting for";
    static const QString path_start_tag = "'s lock in";
    static const QString end_of_line = "\n";

    int user_start( buffer.indexOf( user_start_tag ) );
    if( user_start < 0 ) return false;
    user_start += user_start_tag.length();

    int path_start( buffer.indexOf( path_start_tag, user_start ) );
    if( path_start < 0 ) return false;

    // retrieve user
    QString user( buffer.mid( user_start, path_start - user_start ).trimmed() );
    path_start += path_start_tag.length();

    int path_end( buffer.indexOf( end_of_line, path_start ) );
    if( path_end < 0 ) return false;

    // parse path
    QString path( buffer.mid( path_start, path_end - path_start ).trimmed() );

    // emit message
    {
        QString buffer;
        QTextStream(&buffer) << "Found lock from " << user << " in " << path << endl;
        emit messageAvailable( buffer, 0 , QColor( "#aa0000") );
    }

    // open dialog
    if( !ignoreLocks_ )
    {
        // set ignore_locks to true (to avoid multiple dialogs to be oppened)
        // opens CvsLockDialog.
        // If dialog is accepted, abort all actions.
        ignoreLocks_ = true;
        if( CvsLockDialog( qApp->activeWindow(), user, path ).exec() == QDialog::Accepted ) abort();
    }

    return true;

}

//___________________________________________________________________
void CvsInterface::_logCommandAvailable( void )
{

    // need to check local vs absolute names

    Debug::Throw( "CvsInterface::_logCommandAvailable.\n" );

    // parsing strings
    static const QRegExp separator( "\\s(-{28,})|(={28,})\\s" );
    static const QString revision_tag( "revision" );
    static const QString date_tag( "date:" );
    static const QString author_tag( "author:" );
    static const QString tag_end( ";" );

    // emit full buffer
    emit messageAvailable( stdoutBuffer_.text().mid( stdoutBuffer_.position() ), FORMAT::ITALIC, QColor(), HIGH );

    // look for first separator in full buffer. It is needed because at least one separator is needed
    // to start the revisions
    int first_separator( stdoutBuffer_.text().indexOf( separator ) );
    if( first_separator < 0 ) return;

    int start = std::max( stdoutBuffer_.position(), first_separator + separator.matchedLength() );

    Revision revision;
    Revision::List revisions;
    while(1)
    {

        // allocate current revision
        Revision revision;

        // look for revision
        int revision_start = stdoutBuffer_.text().indexOf( revision_tag, start );
        if( revision_start < 0 ) break;
        revision_start += revision_tag.length();

        // look for end of line
        int end_of_line = stdoutBuffer_.text().indexOf( "\n", revision_start );
        if( end_of_line < 0 ) break;

        // parse revision name
        QString revision_name = stdoutBuffer_.text().mid( revision_start, end_of_line - revision_start ).trimmed();
        revision.setName( revision_name );

        // look for date
        int date_start = stdoutBuffer_.text().indexOf( date_tag, end_of_line );
        if( date_start < 0 ) break;
        date_start += date_tag.length();

        int date_end = stdoutBuffer_.text().indexOf( tag_end, date_start );
        if( date_end < 0 ) break;

        // parse date
        QString date_string = stdoutBuffer_.text().mid( date_start, date_end-date_start ).trimmed();
        revision.setTimeStamp( date_string );

        // look for author
        int author_start = stdoutBuffer_.text().indexOf( author_tag, date_end );
        if( author_start < 0 ) break;
        author_start += author_tag.length();

        int author_end = stdoutBuffer_.text().indexOf( tag_end, author_start );
        if( author_end < 0 ) break;

        // parse author
        QString author = stdoutBuffer_.text().mid( author_start, author_end - author_start ).trimmed();
        revision.setAuthor( author );

        // look for next end of line to start reading comments
        int comments_start = stdoutBuffer_.text().indexOf( "\n", author_end );
        if( comments_start < 0 ) break;
        comments_start++;

        // look for next separator for end of comments
        int comments_end = stdoutBuffer_.text().indexOf( separator, comments_start );
        if( comments_end < 0 ) break;

        // parse comments
        QString comments = stdoutBuffer_.text().mid( comments_start, comments_end - comments_start );
        revision.setComments( comments );

        // store revision and emit message
        revisions.push_back( revision );
        QString message( revision.name() + " date: " + revision.timeStamp().toString() + " author: " + revision.author() + "\n" );
        emit messageAvailable( message, 0, QColor( "#0000aa") );

        // update flags
        start = comments_end + separator.matchedLength();

    }

    // update position in buffer
    stdoutBuffer_.flush( start );

    // send found revisions and return
    if( !revisions.empty() );
    emit revisionsAvailable( commands_.front().argument(0), revisions );
    return;

}

//___________________________________________________________________
void CvsInterface::_statusCommandAvailable( void )
{

    Debug::Throw( "CvsInterface::_statusCommandAvailable.\n" );

    // map cvs tags versus fileInfo status
    static QHash< QString, FileInfo::Status > status_map;
    if( status_map.empty() )
    {
        status_map["Up-to-date"] = FileInfo::UpToDate;
        status_map["Locally Modified"] = FileInfo::LocallyModified;
        status_map["Locally Added"] = FileInfo::LocallyAdded;
        status_map["Locally Removed"] = FileInfo::LocallyRemoved;
        status_map["Needs Patch"] = FileInfo::RemotelyModified;
        status_map["Entry Invalid"] = FileInfo::RemotelyRemoved;
        status_map["Needs Merge"] = FileInfo::Merge;
        status_map["Needs Checkout"] = FileInfo::RemotelyAdded;
        status_map["File had Conflicts on Merge"] = FileInfo::Conflict;
    }

    QRegExp file_tag( "File: (no file )?" );
    static const QString status_tag = "Status: ";
    static const QString end_of_line = "\n";

    // emit full buffer
    emit messageAvailable( stdoutBuffer_.text().mid( stdoutBuffer_.position() ), FORMAT::ITALIC, QColor(), HIGH );

    // list of loaded information
    QString path( commands_.front().argument(0) );
    FileInfo::List files;

    // parse buffer
    while(1)
    {

        // look for file tag
        int file_pos = stdoutBuffer_.text().indexOf( file_tag, stdoutBuffer_.position() );
        if( file_pos < 0 ) break;

        int position( file_pos + file_tag.matchedLength() );

        // look for status tag
        int status_pos = stdoutBuffer_.text().indexOf( status_tag, position );
        if( status_pos < 0 ) break;

        // look for next end of line
        int end_pos = stdoutBuffer_.text().indexOf( end_of_line, status_pos );
        if( end_pos < 0 ) break;

        // retrieve file and status
        QString file = QString( stdoutBuffer_.text() ).mid( position, status_pos - position ).trimmed();
        QString status = QString( stdoutBuffer_.text() ).mid( status_pos + status_tag.size(), end_pos - status_pos - status_tag.size() ).trimmed();

        QHash< QString, FileInfo::Status >::iterator status_iter( status_map.find( status ) );
        files.push_back( FileInfo( File(file).addPath( path ), FileInfo::Document, status_iter.value() ) );

        QString buffer;
        QTextStream(&buffer) << status << " - " << file << endl;
        emit messageAvailable( buffer, 0, QColor( QColor( "#0000aa") ) );

        // update position for next bunch of available data
        stdoutBuffer_.flush( end_pos + end_of_line.size() );

    }

    emit statusAvailable( path, files );
    return;

}

//___________________________________________________________________
void CvsInterface::_fakeUpdateCommandAvailable( void )
{
    Debug::Throw( "CvsInterface::_fakeUpdateCommandAvailable.\n" );

    static const QRegExp tag( "\\b(U|P|C|M)\\b" );
    static const QString end_of_line( "\n" );

    static QHash<QString, QString> tag_map;
    static QHash<QString, FileInfo::Status> status_map;

    if( tag_map.empty() )
    {
        tag_map["U"] = "Remotely Added";
        tag_map["M"] = "Locally Modified";
        tag_map["P"] = "Remotely Modified";
        tag_map["C"] = "Conflict";
    }

    if( status_map.empty() )
    {
        status_map["U"] = FileInfo::RemotelyModified;
        status_map["P"] = FileInfo::RemotelyModified;
        status_map["M"] = FileInfo::LocallyModified;
        status_map["C"] = FileInfo::Conflict;
    }

    // emit full buffer
    emit messageAvailable( stdoutBuffer_.text().mid( stdoutBuffer_.position() ), FORMAT::ITALIC, QColor(), HIGH );

    // list of loaded information
    QString path( commands_.front().argument(0) );
    FileInfo::List files;
    while(1)
    {

        int start_pos( tag.indexIn( stdoutBuffer_.text().left( stdoutBuffer_.position() ) ) );
        if( start_pos < 0 ) break;

        int length( tag.matchedLength() );

        // parse tag and lookup in conversion table
        QString tag_match = stdoutBuffer_.text().mid( start_pos, length );

        QHash<QString, FileInfo::Status>::iterator status_iter( status_map.find( tag_match ) );
        assert( status_iter != status_map.end() );
        FileInfo::Status status( status_iter.value() );

        // convert tag to readable string
        QHash<QString, QString>::iterator tag_iter( tag_map.find( tag_match ) );
        assert( tag_iter != tag_map.end() );
        tag_match = tag_iter.value();

        // look for file
        stdoutBuffer_.flush( start_pos + length );
        int end( stdoutBuffer_.text().indexOf( end_of_line, stdoutBuffer_.position() ) );
        if( end < 0 ) break;

        QString file = stdoutBuffer_.text().mid( stdoutBuffer_.position(), end-stdoutBuffer_.position() ).trimmed();
        files.push_back( FileInfo( File(file).addPath( path ), FileInfo::Document, status ) );

        QString buffer;
        QTextStream(&buffer) << tag_match << " - " << file << endl;
        emit messageAvailable( buffer, 0, QColor( QColor( "#0000aa") ) );

        stdoutBuffer_.flush( end );

    }

    emit statusAvailable( commands_.front().argument(0), files );
    return;
}

//___________________________________________________________________
void CvsInterface::_revisionCheckOutCommandCompleted( void )
{
    Debug::Throw( "CvsInterface::_revisionCheckOutCommandCompleted.\n" );

    // retrieve pending command
    CvsCommand& command( commands_.front() );

    // generate destination filename from source and revision name
    QString destination( command.argument(0) );
    QString rev = QString( command.argument(1) ).replace( " ", "_" );
    destination = QFileInfo( destination ).canonicalPath() + "/" + QFileInfo( destination ).baseName() + "-" + rev + "." + QFileInfo( destination ).completeSuffix();

    // write buffer to destination file
    QFile out( destination );
    if( !out.open( QIODevice::WriteOnly ) )
    {
        QString buffer;
        QTextStream(&buffer) << "unable to write to " << destination << ". <cvs up> canceled.";
        InformationDialog( qApp->activeWindow(), buffer ).exec();
        return;
    }

    // retrieve all stdout and put into stream
    out.write( stdoutBuffer_.text().toAscii() );
    out.close();

    QString buffer;
    QTextStream(&buffer) << destination << " created" << endl;
    emit messageAvailable( buffer, 0, QColor( "#0000aa") );
    emit revisionAvailable( command.argument(0), destination );
    emit scratchFileAdded( File( destination ) );

}

//___________________________________________________________________
void CvsInterface::_checkOutCommandAvailable( void )
{
    Debug::Throw( "CvsInterface::_checkOutCommandAvailable.\n" );

    static const QRegExp tag( "\\b(U)\\b" );
    static const QString end_of_line( "\n" );
    static QHash<QString, QString> tag_map;

    if( tag_map.empty() )
    {
        tag_map["U"] = "Added";
    }

    // emit full buffer
    emit messageAvailable( stdoutBuffer_.text().mid( stdoutBuffer_.position() ), FORMAT::ITALIC, QColor(), HIGH );

    // list of loaded information
    FileInfo::List files;
    while(1)
    {

        // look for tag
        int start_pos( tag.indexIn( stdoutBuffer_.text().left( stdoutBuffer_.position() ) ) );
        if( start_pos < 0 ) break;
        int length( tag.matchedLength() );
        if( length > 1 )
        {
            start_pos ++;
            length --;
        }

        // parse tag and lookup in conversion table
        QString tag = stdoutBuffer_.text().mid( start_pos, length );
        QHash<QString, QString>::iterator iter( tag_map.find( tag ) );
        assert( iter != tag_map.end() );
        tag = iter.value();

        // look for file
        stdoutBuffer_.flush( start_pos + length );
        int end( stdoutBuffer_.text().indexOf( end_of_line, stdoutBuffer_.position() ) );
        if( end < 0 ) break;

        QString file = stdoutBuffer_.text().mid( stdoutBuffer_.position(), end-stdoutBuffer_.position() ).trimmed();
        files.push_back( FileInfo( file, FileInfo::Document, FileInfo::UpToDate ) );

        QString buffer;
        QTextStream( &buffer ) << tag << " - " << file << endl;
        emit messageAvailable( buffer, 0, QColor( QColor( "#0000aa") ) );

        stdoutBuffer_.flush( end + end_of_line.size() );

    }

    emit statusAvailable( commands_.front().argument(0), files );
    return;

};

//___________________________________________________________________
void CvsInterface::_updateCommandAvailable( void )
{
    Debug::Throw( "CvsInterface::_updateCommandAvailable.\n" );
    static const QRegExp tag( "\\b(U|P|M|C|A|R)\\b" );
    static const QString end_of_line( "\n" );
    static QHash<QString, QString> tag_map;
    static QHash<QString, FileInfo::Status> status_map;

    if( tag_map.empty() )
    {
        tag_map["U"] = "Updated";
        tag_map["P"] = "Updated";
        tag_map["M"] = "Locally Modified";
        tag_map["C"] = "Conflict";
        tag_map["A"] = "Locally Added";
        tag_map["R"] = "Locally Removed" ;
    }

    // maps the tag to the status of the file *after* the update has been performed
    if( status_map.empty() )
    {
        status_map["U"] = FileInfo::UpToDate;
        status_map["P"] = FileInfo::UpToDate;
        status_map["M"] = FileInfo::LocallyModified;
        status_map["C"] = FileInfo::Conflict;
        status_map["A"] = FileInfo::LocallyAdded;
        status_map["R"] = FileInfo::LocallyRemoved ;
    }

    // emit full buffer
    emit messageAvailable( stdoutBuffer_.text().mid( stdoutBuffer_.position() ), FORMAT::ITALIC, QColor(), HIGH );

    FileInfo::List files;
    while(1)
    {

        // look for tag
        int start_pos( tag.indexIn( stdoutBuffer_.text().left( stdoutBuffer_.position() ) ) );
        if( start_pos < 0 ) break;

        int length( tag.matchedLength() );

        // parse tag and lookup in conversion table
        QString tag = stdoutBuffer_.text().mid( start_pos, length );
        QHash<QString, FileInfo::Status>::iterator status_iter( status_map.find( tag ) );
        FileInfo::Status status( status_iter.value() );

        // convert tag to readable string
        QHash<QString, QString>::iterator tag_iter( tag_map.find( tag ) );
        assert( tag_iter != tag_map.end() );
        tag = tag_iter.value();

        // look for file
        stdoutBuffer_.flush( start_pos + length );
        int end( stdoutBuffer_.text().indexOf( end_of_line, stdoutBuffer_.position() ) );
        if( end < 0 ) break;

        QString file = stdoutBuffer_.text().mid( stdoutBuffer_.position(), end-stdoutBuffer_.position() ).trimmed();
        files.push_back( FileInfo( file, FileInfo::Document, status ) );

        QString buffer;
        QTextStream(&buffer) << tag << " - " << file << endl;
        emit messageAvailable( buffer, 0, QColor( QColor( "#0000aa") ) );

        stdoutBuffer_.flush( end + end_of_line.size() );

    }

    emit statusAvailable( commands_.front().argument(0), files );
    return;

}

//___________________________________________________________________
void CvsInterface::_commitCommandAvailable( void )
{
    Debug::Throw( "CvsInterface::_commitCommandAvailable.\n" );

    // retrieve pending command
    CvsCommand& command( commands_.front() );

    // emit full buffer
    emit messageAvailable( stdoutBuffer_.text().mid( stdoutBuffer_.position() ), FORMAT::ITALIC, QColor(), HIGH );

    // look for committed files
    static const QRegExp commit_tag( "(Checking in)|(Removing)" );
    static const QString end_of_line( ";\n" );
    FileInfo::List files;
    while(1)
    {

        // look for tag
        int start_pos( commit_tag.indexIn( stdoutBuffer_.text().left( stdoutBuffer_.position() ) ) );
        if( start_pos < 0 ) break;

        // look for file
        stdoutBuffer_.flush( start_pos + commit_tag.matchedLength() );
        int end( stdoutBuffer_.text().indexOf( end_of_line, stdoutBuffer_.position() ) );
        if( end < 0 ) break;

        QString file = stdoutBuffer_.text().mid( stdoutBuffer_.position(), end-stdoutBuffer_.position() ).trimmed();
        files.push_back( FileInfo( file, FileInfo::Document, FileInfo::UpToDate ) );

        QString buffer;
        QTextStream(&buffer) << "Committed - " << File( file ).addPath( command.argument(0) ) << endl;
        emit messageAvailable( buffer, 0, QColor( QColor( "#0000aa") ) );

        stdoutBuffer_.flush( end + end_of_line.size() );

    }

    emit statusAvailable( commands_.front().argument(0), files );
    return;

}

//___________________________________________________________________
void CvsInterface::_commitErrorAvailable( void )
{
    Debug::Throw( "CvsInterface::_commitErrorAvailable.\n" );

    // here you could parse the output and log important informations
    // look for commit aborted
    static const QString error_tag( "cvs [commit aborted]" );
    static const QString previous_tag( "cvs commit:" );

    // look for error message
    int error_pos( stderrBuffer_.text().indexOf( error_tag, stderrBuffer_.position() ) );
    if( error_pos < 0 ) return;

    // look for previous tag
    QString buffer;
    int previous_pos( stderrBuffer_.text().lastIndexOf( previous_tag, error_pos ) );
    if( previous_pos < 0 )
    {
        QTextStream( &buffer )
            << "Commit aborted for an unknown reason." << endl
            << "Retry with log frame visible and maximum verbosity" << endl
            << "to get more information.";

    } else {

        QTextStream( &buffer )
            << "Commit aborted for the following reason:" << endl
            << stderrBuffer_.text().mid( previous_pos + previous_tag.length(), error_pos - (previous_pos + previous_tag.length()) ).trimmed() << endl
            << "Fix the error and retry." << endl;

    }

    InformationDialog(  qApp->activeWindow(), buffer ).exec();
    emit messageAvailable( buffer, FORMAT::ITALIC, QColor( QColor( "#aa0000") ) );
    stderrBuffer_.flush( error_pos + error_tag.length() );

    return;

}

//___________________________________________________________________
void CvsInterface::_commitCommandCompleted( void )
{
    Debug::Throw( "CvsInterface::_commitCommandCompleted.\n" );

    // retrieve pending command
    CvsCommand& command( commands_.front() );

    // retrieve temporary file and try remove
    File tmp( command.argument(1) );
    if( tmp.exists() ) tmp.remove();

}

//___________________________________________
void CvsInterface::_resetProcess( void )
{
    if( process_ )
    {
        process_->deleteLater();
        process_ = 0;
    }
}

//____________________________________________________________
CustomProcess& CvsInterface::_newProcess( void )
{

    // delete old process
    _resetProcess();

    // create new one
    process_ = new CustomProcess( this );
    connect( &_process(), SIGNAL( finished( int, QProcess::ExitStatus ) ), this, SLOT( _commandCompleted( int, QProcess::ExitStatus ) ) );
    connect( &_process(), SIGNAL( error(  QProcess::ProcessError ) ), SLOT( _error( QProcess::ProcessError ) ) );
    connect( &_process(), SIGNAL( readyReadStandardOutput() ), SLOT( _readStdout() ) );
    connect( &_process(), SIGNAL( readyReadStandardError() ), SLOT( _readStderr() ) );

    return _process();

}
