// $Id: FileInfo.cpp,v 1.16 2012/09/22 07:53:45 hpereira Exp $

/******************************************************************************
*
* Copyright (C) 2002 Hugo PEREIRA <mailto: hugo.pereira@free.fr>
*
* This is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
*
* This software is distributed in the hope that it will be useful, but WITHOUT
* ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License along with
* software; if not, write to the Free Software Foundation, Inc., 59 Temple
* Place, Suite 330, Boston, MA  02111-1307 USA
*
*
*******************************************************************************/

#include "FileInfo.h"
#include "Debug.h"
#include "TimeStamp.h"

#include <cassert>

//! Some Xml definitions
namespace XML
{
    static const QString CREATION( "Creation" );
    static const QString STATUS( "Status" );
}

//________________________________________________________________
FileInfo::FileInfo( File file, unsigned int type ):
    BaseFileInfo( file, type ),
    Counter( "FileInfo" ),
    dirty_( true ),
    status_( Unknown ),
    creation_( TimeStamp::now() )
{}

//________________________________________________________________
FileInfo::FileInfo( File file, unsigned int type, Status status ):
    BaseFileInfo( file, type ),
    Counter( "FileInfo" ),
    dirty_( true ),
    status_( status ),
    creation_( TimeStamp::now() )
{}

//________________________________________________________________
FileInfo::FileInfo( const QDomElement& element ):
    BaseFileInfo( element ),
    Counter( "FileInfo" ),
    dirty_( true ),
    status_( Unknown ),
    creation_(0)
{

    Debug::Throw( "FileInfo::FileInfo (dom).\n" );

    // parse attributes
    QDomNamedNodeMap attributes( element.attributes() );
    for( unsigned int i=0; i<attributes.length(); i++ )
    {
        QDomAttr attribute( attributes.item( i ).toAttr() );
        if( attribute.isNull() ) continue;
        else if( attribute.name() == XML::STATUS ) setStatus( attribute.value().toUInt() );
        else if( attribute.name() == XML::CREATION ) setCreation( attribute.value().toInt() );
    }

}

//________________________________________________________________
QDomElement FileInfo::domElement( QDomDocument& document ) const
{
    Debug::Throw( "FileInfo::DomElement.\n" );
    QDomElement out( BaseFileInfo::domElement( document ) );
    out.setAttribute( XML::STATUS, QString().setNum( status() ) );
    out.setAttribute( XML::CREATION, QString().setNum( created() ) );
    return out;
}

//________________________________________________________________
void FileInfo::update( void )
{

    if( !dirty() ) return;

    Debug::Throw() << "FileInfo::update - file: " << file() << endl;
    if( file().isEmpty() )
    {

        setUser( QString() );
        setGroup( QString() );
        setSize( 0 );
        setLastModified( TimeStamp() );
        setPermissions( QFile::Permissions() );

    } else {

        setSize( file().fileSize() );
        setUser( file().userName() );
        setGroup( file().groupName() );
        setLastModified( file().lastModified() );
        setPermissions( file().permissions() );

    }

    setDirty( false );

}

//________________________________________________________________
void FileInfo::updateFrom( const FileInfo& info )
{

    if( !dirty() ) return;

    Debug::Throw() << "FileInfo::updateFrom - file: " << file() << endl;
    assert( file() == info.file() );
    assert( !info.dirty() );

    setUser( info.user() );
    setGroup( info.group() );
    setSize( info.size() );
    setLastModified( info.lastModified() );
    setPermissions( info.permissions() );

    setDirty( false );

}
