#ifndef FileInfo_h
#define FileInfo_h

// $Id: FileInfo.h,v 1.20 2012/09/22 07:53:45 hpereira Exp $

/******************************************************************************
*
* Copyright (C) 2002 Hugo PEREIRA <mailto: hugo.pereira@free.fr>
*
* This is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
*
* This software is distributed in the hope that it will be useful, but WITHOUT
* Any WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License along with
* software; if not, write to the Free Software Foundation, Inc., 59 Temple
* Place, Suite 330, Boston, MA  02111-1307 USA
*
*
*******************************************************************************/

#include "BaseFileInfo.h"
#include "QOrderedSet.h"

#include <QtXml/QDomDocument>
#include <QtXml/QDomElement>

#include <cassert>

//! store a file and needed informations for Ftp transfer
class FileInfo: public BaseFileInfo, public Counter
{

    public:

    //! cvs status
    enum Status
    {

        //! unknown
        Unknown = 1<<0,

        //! up to date
        UpToDate = 1<<1,

        //! modified locally
        LocallyModified = 1<<2,

        //! added
        LocallyAdded = 1<<3,

        //! removed (from cvs up)
        LocallyRemoved = 1<<4,

        //! modified remotely
        RemotelyModified = 1<<5,

        //! remotely added (from cvs up)
        RemotelyAdded = 1<<6,

        //! remotely removed
        RemotelyRemoved = 1<<7,

        //! need Merge
        Merge = 1<<8,

        //! Conflict (from cvs up)
        Conflict = 1<<9,

    };

    //! constructor
    FileInfo( File file = File(), unsigned int type = None );

    //! constructor
    FileInfo( File file, unsigned int type, Status status );

    //! constructor from DOM element
    FileInfo( const QDomElement& );

    //! dump to dom element
    QDomElement domElement( QDomDocument& parent ) const;

    //!@name accessors
    //@{

    //! creation
    const time_t created( void ) const
    { return creation_; }

    //! true when file and local information are not synchronized
    const bool& dirty( void ) const
    { return dirty_; }

    //! get status
    unsigned int status( void ) const
    { return status_; }

    //@}

    //!@name modifiers
    //@{

    //! file
    virtual void setFile( const File& file )
    {
        BaseFileInfo::setFile( file );
        setDirty( true );
    }

    //! dirty
    void setDirty( bool value )
    { dirty_ = value; }

    //! type
    virtual void setType( unsigned int type )
    {
        assert( !( type&(~Any) ) );
        BaseFileInfo::setType( type );
    }

    //! set status
    void setStatus( unsigned int status )
    { status_ = status; }

    //! time stamp
    void setCreation( time_t value )
    { creation_ = value; }

    //! update internal storage from file
    void update( void );

    //! update internal storage from existing file info
    void updateFrom( const FileInfo& );

    //@}

    //! needed to sort fileInfo by creation timeStamp
    class FirstCreatedFTor
    {
        public:

        //! sort operator
        bool operator() (const FileInfo& first, const FileInfo& second ) const
        { return first.created() < second.created(); }


    };

    //! used to count files that do not match a given type
    class DifferentTypeFTor
    {
        public:

        //! constructor
        DifferentTypeFTor( const unsigned int& type ):
            type_( type )
        {}

        //! predicate
        /*! the construct (info.type()&type_) == type_ enables to use bit pattern for a type */
        bool operator() (const FileInfo& info ) const
        { return !(info.type()&type_); }

        private:

        //! predicted type
        unsigned int type_;

    };

    //! needed to retrieve items that are documents
    class ExistsFTor
    {

        public:

        //! constructor
        ExistsFTor( File path ):
            path_( path )
        {}

        //! prediction
        bool operator() ( const FileInfo& info ) const
        { return !( info.type() & FileInfo::Navigator ) && info.file().addPath( path_ ).exists(); }

        private:

        //! base path
        File path_;

    };

    //! needed to retrieve items that are documents
    class IsExistingDocumentFTor
    {

        public:

        //! constructor
        IsExistingDocumentFTor( File path ):
            path_( path )
        {}

        //! prediction
        bool operator() ( const FileInfo& info ) const
        { return  ( info.type() & FileInfo::Document ) && info.file().addPath( path_ ).exists(); }

        private:

        //! base path
        File path_;

    };

    //@}

    //! list
    typedef BaseFileInfo::BaseList<FileInfo> List;

    //! set
    typedef QOrderedSet<FileInfo> Set;

    private:

    //! true if file and information are not synchronized
    bool dirty_;

    //! file type
    unsigned int status_;

    //! creation time
    time_t creation_;

};

#endif
