// $Id: FileInfoModel.cpp,v 1.33 2012/09/24 11:41:18 hpereira Exp $

/******************************************************************************
*
* Copyright (C) 2002 Hugo PEREIRA <mailto: hugo.pereira@free.fr>
*
* This is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
*
* This software is distributed in the hope that it will be useful, but WITHOUT
* Any WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License along with
* software; if not, write to the Free Software Foundation, Inc., 59 Temple
* Place, Suite 330, Boston, MA  02111-1307 USA
*
*
*******************************************************************************/

#include "CustomPixmap.h"
#include "FileInfoModel.h"
#include "Icons.h"
#include "Singleton.h"
#include "TimeStamp.h"
#include "XmlOptions.h"

#include <QtGui/QIcon>

//_______________________________________________________________
FileInfoModel::FileInfoModel( QObject* parent ):
    BaseFileInfoModel<FileInfo>( parent ),
    Counter( "FileInfoModel" )
{
    Debug::Throw( "FileInfoModel::FileInfoModel.\n" );

    connect( Singleton::get().application(), SIGNAL( configurationChanged() ), SLOT( _updateConfiguration() ) );
    _updateConfiguration();

}

//__________________________________________________________________
QVariant FileInfoModel::data( const QModelIndex& index, int role ) const
{

    // check index, role and column
    if( !index.isValid() ) return QVariant();

    // return icon associated to file
    if( showIcons() && role == Qt::DecorationRole && index.column() == 0 )
    {
        const FileInfo& info( get()[index.row()] );
        int type( info.type() );
        type &= (FileInfo::Document|FileInfo::Folder|FileInfo::Link|FileInfo::Hidden|FileInfo::Navigator);
        type |= info.status()<<FileInfo::Shift;
        QIcon icon( _icon( type ) );
        return icon;
    }

    // fallback to parent
    return BaseFileInfoModel<FileInfo>::data( index, role );

}

//_________________________________________________________
void FileInfoModel::_updateConfiguration( void )
{

    Debug::Throw( "FileInfoModel::_updateConfiguration.\n" );

    // install pixmaps
    icons_.clear();
    _installIcons();
    reset();

}

//_____________________________________________________________________
void FileInfoModel::_installIcons( void )
{

    Debug::Throw( "FileInfoModel::_installIcons.\n" );

    if( !icons_.empty() ) return;

    // pixmap size
    unsigned int pixmap_size = XmlOptions::get().get<unsigned int>( "LIST_ICON_SIZE" );
    QSize size( pixmap_size, pixmap_size );
    QSize scale(size*0.9);

    // type icons
    typedef QHash< int, QString > IconNames;

    // small icons
    IconNames typeNames;
    typeNames[FileInfo::Document] = ICONS::DOCUMENT;
    typeNames[FileInfo::Folder] = ICONS::FOLDER;
    typeNames[FileInfo::Navigator] = ICONS::PARENT_DIRECTORY;
    CustomPixmap link = CustomPixmap().find( ICONS::LINK );
    for( IconNames::iterator iter = typeNames.begin(); iter != typeNames.end(); ++iter )
    {

        unsigned int type( iter.key()|(FileInfo::Unknown<<FileInfo::Shift) );
        {
            // normal
            const CustomPixmap pixmap( CustomPixmap()
                .empty( size )
                .merge( CustomPixmap().find( iter.value() )
                .scaled( scale, Qt::KeepAspectRatio, Qt::SmoothTransformation ), CustomPixmap::CENTER ) );
            icons_[type] = pixmap;
            icons_[type|FileInfo::Hidden] = pixmap.transparent( 0.7 );
        }

        {
            // link
            const CustomPixmap pixmap( CustomPixmap()
                .empty( size )
                .merge( CustomPixmap().find( iter.value() )
                .merge( link, CustomPixmap::BOTTOM_LEFT )
                .scaled( scale, Qt::KeepAspectRatio, Qt::SmoothTransformation ), CustomPixmap::CENTER ) );

            icons_[type|FileInfo::Link] = pixmap;
            icons_[type|FileInfo::Link|FileInfo::Hidden] = pixmap.transparent( 0.7 );
        }

    }

    // large icons
    IconNames typeNamesLarge;
    typeNamesLarge[FileInfo::Document] = ICONS::DOCUMENT_LARGE;
    typeNamesLarge[FileInfo::Folder] = ICONS::FOLDER_LARGE;
    CustomPixmap linkOverlayLarge = CustomPixmap().find( ICONS::LINK_LARGE );
    for( IconNames::iterator iter = typeNamesLarge.begin(); iter != typeNamesLarge.end(); ++iter )
    {
        unsigned int type( iter.key()|(FileInfo::Unknown<<FileInfo::Shift) );

        const CustomPixmap pixmap( CustomPixmap().find( iter.value() ) );
        icons_[type].addPixmap( pixmap );
        icons_[type|FileInfo::Hidden].addPixmap( pixmap.transparent(0.7) );

        const CustomPixmap link( pixmap.merge( linkOverlayLarge, CustomPixmap::BOTTOM_RIGHT ) );
        icons_[type|FileInfo::Link].addPixmap( link );
        icons_[type|FileInfo::Link|FileInfo::Hidden].addPixmap( link.transparent( 0.7 ) );
    }

    // status icons
    IconNames statusNames;
    statusNames[FileInfo::Unknown] = ICONS::UNKNOWN;
    statusNames[FileInfo::UpToDate] = ICONS::UP_TO_DATE;
    statusNames[FileInfo::LocallyModified] = ICONS::LOCALLY_MODIFIED;
    statusNames[FileInfo::LocallyAdded] = ICONS::LOCALLY_ADDED;
    statusNames[FileInfo::LocallyRemoved] = ICONS::LOCALLY_REMOVED;
    statusNames[FileInfo::RemotelyModified] = ICONS::REMOTELY_MODIFIED;
    statusNames[FileInfo::RemotelyAdded] = ICONS::REMOTELY_ADDED;
    statusNames[FileInfo::RemotelyRemoved] = ICONS::REMOTELY_REMOVED;
    statusNames[FileInfo::Merge] = ICONS::MERGE;
    statusNames[FileInfo::Conflict] = ICONS::CONFLICT;

    CustomPixmap base(
        CustomPixmap().find( typeNames[FileInfo::Document] )
        .scaled( size, Qt::KeepAspectRatio, Qt::SmoothTransformation ) );

    CustomPixmap baseLarge( CustomPixmap().find( typeNamesLarge[FileInfo::Document] ) );
    for( IconNames::iterator iter = statusNames.begin(); iter != statusNames.end(); ++iter )
    {
        const int type( FileInfo::Document | (iter.key() << FileInfo::Shift ) );
        {
            const CustomPixmap pixmap( CustomPixmap()
                .empty( size )
                .merge( base
                .merge(CustomPixmap()
                .find( iter.value() )
                .scaled( 0.6*size, Qt::KeepAspectRatio, Qt::SmoothTransformation ) )
                .scaled( scale, Qt::KeepAspectRatio, Qt::SmoothTransformation ), CustomPixmap::CENTER ) );

            icons_[type] =  pixmap;
            icons_[type|FileInfo::Hidden] = pixmap.transparent( 0.7 );
        }

        {

            const CustomPixmap pixmap( baseLarge.merge( CustomPixmap()
                .find( iter.value() )
                .scaled( 0.4*baseLarge.size(), Qt::KeepAspectRatio, Qt::SmoothTransformation ) ) );
            icons_[type].addPixmap( pixmap );
            icons_[type|FileInfo::Hidden].addPixmap( pixmap.transparent( 0.7 ) );
        }
    }

    return;
}
