// $Id: FileSelectionFrame.cpp,v 1.142 2012/09/24 11:41:36 hpereira Exp $

/******************************************************************************
*
* Copyright (C) 2002 Hugo PEREIRA <mailto: hugo.pereira@free.fr>
*
* This is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
*
* This software is distributed in the hope that it will be useful, but WITHOUT
* Any WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License along with
* software; if not, write to the Free Software Foundation, Inc., 59 Temple
* Place, Suite 330, Boston, MA  02111-1307 USA
*
*
*******************************************************************************/

#include "FileSelectionFrame.h"

#include "AnimatedTreeView.h"
#include "Command.h"
#include "ColumnSelectionMenu.h"
#include "ColumnSortingMenu.h"
#include "ContextMenu.h"
#include "CvsAddDialog.h"
#include "CvsCheckOutDialog.h"
#include "CvsCommitDialog.h"
#include "CvsRemoveDialog.h"
#include "Icons.h"
#include "IconEngine.h"
#include "IconView.h"
#include "InformationDialog.h"
#include "PathEditor.h"
#include "RemoveFilesDialog.h"
#include "Singleton.h"
#include "TextEditor.h"
#include "BaseFileInfoToolTipWidget.h"
#include "Util.h"
#include "XmlOptions.h"

#include <QtGui/QLayout>
#include <QtCore/QDir>


//_____________________________________________
FileSelectionFrame::FileSelectionFrame( QWidget *parent ):
    QWidget( parent ),
    Counter( "FileSelectionFrame" ),
    model_( this ),
    selectionModel_( &model_, this ),
    list_( 0x0 ),
    viewModeOptionName_( "VIEW_MODE" ),
    viewMode_( ICONS ),
    contextMenu_( 0x0 ),
    columnSelectionMenu_( 0x0 ),
    columnSortingMenu_( 0x0 ),
    interface_( this ),
    fileSystemWatcher_( this ),
    showNavigator_( true )
{

    Debug::Throw( "FileSelectionFrame::FileSelectionFrame.\n" );
    QVBoxLayout *layout = new QVBoxLayout();
    layout->setSpacing(0);
    layout->setMargin(0);
    setLayout( layout );

    // install actions
    _installActions();

    // editor
    editor_ = new PathEditor( this );

    // initialize local home path
    editor_->setHomePath( Util::home() );

    // initialize local root path list
    File::List rootPathList;
    foreach( const QFileInfo& fileInfo, QDir::drives() )
    { rootPathList << fileInfo.path(); }

    editor_->setRootPathList( rootPathList );

    // tooltip manager
    toolTipWidget_ = new BaseFileInfoToolTipWidget( this );

    // list
    viewMode_ = (FileSelectionFrame::ViewMode) XmlOptions::get().get<unsigned int>( viewModeOptionName_ );
    _setupList();

    connect( &interface_, SIGNAL( statusAvailable( File, FileInfo::List ) ), SLOT( _showStatus( File, FileInfo::List ) ) );
    connect( &interface_, SIGNAL( commandCompleted( CvsCommand ) ), SLOT( commandCompleted( CvsCommand ) ) );

    _updateActions();
    _updateNavigationActions();

    connect( &model_, SIGNAL( layoutAboutToBeChanged() ), SLOT( _storeSelection() ) );
    connect( &model_, SIGNAL( layoutChanged() ), SLOT( _restoreSelection() ) );

    // fileSystem watcher
    connect( &fileSystemWatcher_, SIGNAL( directoryChangedDelayed( const QString& ) ), SLOT( reload( void ) ) );

    // configuration
    connect( Singleton::get().application(), SIGNAL( configurationChanged( void ) ), SLOT( _updateConfiguration( void ) ) );
    _updateConfiguration();

}

//_________________________________________________________
void FileSelectionFrame::setFile( File file )
{

    Debug::Throw() << "FileSelectionFrame::setFile - file: " << file << endl;

    // reset local file
    // store path
    File path( file );

    // split into file and path
    if( !file.isDirectory() )
    {

        path = file.path();
        file = file.localName();

    } else file = File();

    Debug::Throw() << "FileSelectionFrame::setFile - path: " << path << " file: " << file << endl;

    // update local file
    file_ = file;

    // update editor
    editor_->setPath( path, file );

    // update view
    if( !path.isEqual( this->path()  ) )
    {

        Debug::Throw() << "FileSelectionFrame::setFile - current_path: " << FileSelectionFrame::path() << " - Updating" << endl;

        // update fileSystem watcher
        const QStringList directories( fileSystemWatcher_.directories() );
        if( !directories.isEmpty() ) fileSystemWatcher_.removePaths( directories );
        if( !path.isEmpty() ) fileSystemWatcher_.addPath( path );

        // initialize transition
        AnimatedTreeView* treeView( qobject_cast<AnimatedTreeView*>( list_ ) );
        if( treeView ) treeView->initializeAnimation();

        path_ = path;
        list_->selectionModel()->clear();

        _updateActions();
        _updateNavigationActions();

        // save local cache and update
        cache_.merge( localCache_ );
        localCache_ = cache_.findAll( FileSelectionFrame::path() );

        // reload and start transition
        _reload();
        if( treeView ) treeView->startAnimation();

    } else {

        Debug::Throw() << "FileSelectionFrame::setFile - Updating current path." << endl;

        // reload immediately
        // without clearing the list
        _reload( false );

    }

}

//______________________________________________________
void FileSelectionFrame::commandCompleted( CvsCommand command )
{
    Debug::Throw( "FileSelectionFrame::commandCompleted.\n" );
    switch( command.type() )
    {

        case CvsCommand::CHECKOUT:
        _reload( false, false );
        break;

        /*
        for revision checkout there is no need to clear the list, but cvs fake update
        needs to be done to get the proper file status.
        */
        case CvsCommand::CHECKOUT_REVISION:
        _reload( false, true );
        break;

        /*
        for CvsRemove command, one first need to clear the list
        in case some files have been removed (as opposed to renamed)
        from the disk.
        */
        case CvsCommand::REMOVE:
        _reload();
        break;

        default:
        break;
    }

}

//_________________________________________________________
void FileSelectionFrame::clear()
{

    Debug::Throw( "FileSelectionFrame::Clear.\n" );
    model_.clear();
    _setShowNavigator( showNavigator_ && viewMode_ == LIST );

}

//______________________________________________________
void FileSelectionFrame::setViewMode( ViewMode value )
{

    if( viewMode_ == value ) return;
    Debug::Throw( "FileSelectionFrame::setViewMode.\n" );

    viewMode_ = value;
    XmlOptions::get().set<unsigned int>( viewModeOptionName_, viewMode_ );
    _setupList();
    _setShowNavigator( showNavigator_ && viewMode_ == LIST );

    return;

}

//______________________________________________________
FileInfo::List FileSelectionFrame::selection( void ) const
{ return model_.get( list_->selectionModel()->selectedRows() ); }

//_________________________________________________________
bool FileSelectionFrame::hasSelectedDocuments( bool alsoExist ) const
{

    FileInfoModel::List selection( this->selection() );
    return (alsoExist ) ?
        std::find_if( selection.begin(), selection.end(), FileInfo::IsExistingDocumentFTor( path() ) ) != selection.end():
        std::find_if( selection.begin(), selection.end(), FileInfo::SameTypeFTor( FileInfo::Document ) ) != selection.end();

}

//______________________________________________________
void FileSelectionFrame::_setupList( void )
{
    Debug::Throw( "FileSelectionFrame::setupList.\n" );

    // keep track of focus
    const bool hasFocus( list_ && list_->hasFocus() );

    FileInfo::List selection;

    // cleanup
    if( list_ )
    {
        // store selection
        selection = model_.get( list_->selectionModel()->selectedRows() );

        // delete list
        list_->setParent( 0x0 );
        list_->deleteLater();
    }

    // create new list
    if( viewMode_ == ICONS ) list_ = new IconView( this );
    else list_ = new AnimatedTreeView( this );

    layout()->addWidget( list_ );
    list_->setModel( &model_ );
    list_->setSelectionModel( &selectionModel_ );
    list_->setSelectionMode( QAbstractItemView::ContiguousSelection );
    list_->setDragEnabled( false );
    list_->setAcceptDrops( false );
    list_->setDropIndicatorShown( true );

    const QString optionName( "FILE_LIST" );
    if( viewMode_ == LIST ) static_cast<TreeView*>(list_)->setOptionName( optionName );
    else static_cast<IconView*>(list_)->setOptionName( optionName );

    // recreate context menu
    contextMenu_ = new ContextMenu( list_ );
    connect( contextMenu_, SIGNAL( aboutToShow( void ) ), SLOT( _updateMenu( void ) ) );

    // recreate column selection menu and column sorting menu
    if( columnSelectionMenu_ ) columnSelectionMenu_->deleteLater();
    if( columnSortingMenu_ ) columnSortingMenu_->deleteLater();
    if( QTreeView* treeView = qobject_cast<QTreeView*>( list_ ) )
    {

        columnSelectionMenu_ = new ColumnSelectionMenu( this, treeView );
        columnSortingMenu_ = new ColumnSortingMenu( this, treeView->header() );

    } else if( IconView* iconView = qobject_cast<IconView*>( list_ ) ) {

        columnSelectionMenu_ = 0x0;
        columnSortingMenu_ = new ColumnSortingMenu( this, iconView->header() );

    } else {

        columnSelectionMenu_ = 0x0;
        columnSortingMenu_ = 0x0;

    }

    // list connections
    connect( list_->selectionModel(), SIGNAL( selectionChanged(const QItemSelection &, const QItemSelection &) ), SIGNAL( selectionChanged() ) );
    connect( list_->selectionModel(), SIGNAL( selectionChanged(const QItemSelection &, const QItemSelection &) ), SLOT( _updateActions() ) );
    connect( list_->selectionModel(), SIGNAL( currentChanged(const QModelIndex &, const QModelIndex &) ), SLOT( _itemSelected( const QModelIndex& ) ) );
    connect( list_, SIGNAL( activated( const QModelIndex& ) ), SLOT( _itemActivated( const QModelIndex& ) ) );

    if( viewMode_ == ICONS )
    { connect( list_, SIGNAL( hovered( const QModelIndex& ) ), SLOT( _showToolTip( const QModelIndex& ) ) ); }

    // restore focus
    if( hasFocus ) list_->setFocus();

    // restore selection
    if( !selection.isEmpty() )
    {
        list_->selectionModel()->clear();
        foreach( const FileInfo& fileInfo, selection )
        { list_->selectionModel()->select( model_.index( fileInfo ), QItemSelectionModel::Select|QItemSelectionModel::Rows ); }

    }

}

//______________________________________________________
void FileSelectionFrame::_updateConfiguration( void )
{
    Debug::Throw( "FileSelectionFrame::_updateConfiguration.\n" );
    hiddenFilesAction().setChecked( XmlOptions::get().get<bool>( "SHOW_HIDDEN_FILES" ) );
    _toggleShowNavigator( XmlOptions::get().get<bool>( "SHOW_NAVIGATOR" ) );

}

//______________________________________________________
void FileSelectionFrame::_toggleShowHiddenFiles( bool value )
{
    XmlOptions::get().set<bool>( "SHOW_HIDDEN_FILES", value );
    _reload();
}

//______________________________________________________
void FileSelectionFrame::_itemSelected( const QModelIndex& index )
{
    Debug::Throw( "FileSelectionFrame::_itemSelected.\n" );

    // get path
    const File file( index.isValid() ? model_.get(index).file().addPath( path() ):path() );

    // update editor
    if( !file.isDirectory() ) editor_->setPath( file.path(), file.localName() );

    // emit signal
    emit fileSelected( file );
    return;

}

//______________________________________________________
void FileSelectionFrame::_itemActivated( const QModelIndex& index )
{
    Debug::Throw( "FileSelectionFrame::_itemActivated.\n" );

    if( !index.isValid() ) return;

    // retrieve file
    const FileInfo fileInfo( model_.get( index ) );
    if( fileInfo.type() & FileInfo::Folder )
    {

        // make full name
        File path( fileInfo.file() );
        path = path.addPath( FileSelectionFrame::path() );
        setFile( path );

    } else if( fileInfo.type() & FileInfo::Navigator ) parentDirectoryAction().trigger();
    else _edit( fileInfo.file() );

    Debug::Throw( "FileSelectionFrame::_itemActivated. done.\n" );

}


//______________________________________________________
void FileSelectionFrame::_updateMenu( void )
{
    Debug::Throw( "FileSelectionFrame::_updateMenu.\n" );
    contextMenu_->clear();

    // column handling
    if( TreeView* treeView = qobject_cast<TreeView*>( list_ ) )
    {

        contextMenu_->addAction( &treeView->showHeaderAction() );
        if( !treeView->showHeaderAction().isChecked() )
        {
            contextMenu_->addMenu( new ColumnSelectionMenu( this, treeView ) );
            contextMenu_->addMenu( new ColumnSortingMenu( this, treeView->header() ) );
        }

    } else if( IconView* iconView = qobject_cast<IconView*>( list_ ) ) {

        contextMenu_->addMenu( new ColumnSortingMenu( this, iconView->header() ) );

    }

    // edit
    bool needSeparator( true );
    if( editAction().isEnabled() )
    {
        if( needSeparator ) contextMenu_->addSeparator();
        needSeparator = false;
        contextMenu_->addAction( &editAction() );
    }

    if( removeAction().isEnabled() )
    {
        if( needSeparator ) contextMenu_->addSeparator();
        needSeparator = false;
        contextMenu_->addAction( &removeAction() );

    }

    // cvs actions
    contextMenu_->addSeparator();
    contextMenu_->addAction( &cvsCheckOutAction() );
    if( cvsUpdateAction().isEnabled() ) contextMenu_->addAction( &cvsUpdateAction() );
    contextMenu_->addAction( &cvsUpdateDirectoryAction() );

    contextMenu_->addSeparator();
    contextMenu_->addAction( &cvsAddAction() );
    if( cvsRemoveAction().isEnabled() )contextMenu_->addAction( &cvsRemoveAction() );
    if( cvsCommitAction().isEnabled() )contextMenu_->addAction( &cvsCommitAction() );
    contextMenu_->addAction( &cvsCommitDirectoryAction() );

}

//______________________________________________________
void FileSelectionFrame::_updateActions( void )
{
    Debug::Throw( "FileSelectionFrame::_updateActions.\n" );

    // get selected file info
    const FileInfo::List& selection( this->selection() );

    // enable/disable remove action
    if( selection.isEmpty() )
    {
        removeAction().setEnabled( false );
        removeAction().setText( "Delete" );
    } else {
        removeAction().setEnabled( true );
        removeAction().setText( QString( "Delete " ) + selection.description( FileInfo::List::Designate ) );
    }

    // see if selection has existing documents
    FileInfo::List documents;
    FileInfo::List existingSelection;
    FileInfo::List existingDocuments;
    foreach( const FileInfo& fileInfo, selection )
    {
        if( fileInfo.isDocument() ) documents << fileInfo;
        if( fileInfo.file().exists() )
        {
            existingSelection << fileInfo;
            if( fileInfo.isDocument() ) existingDocuments << fileInfo;
        }
    }

    if( existingDocuments.isEmpty() )
    {
        editAction().setEnabled( false );
        editAction().setText( "Edit" );
    } else {
        editAction().setEnabled( true );
        editAction().setText( QString( "Edit " ) + existingDocuments.description( FileInfo::List::Designate ) );
    }

    // see if selection has existing documents/folder
    const bool hasExistingSelection( !existingSelection.isEmpty() );
    cvsAddAction().setEnabled( hasExistingSelection );

    // see if selection has documents
    const bool hasDocument( !documents.isEmpty() );
    cvsRemoveAction().setEnabled( hasDocument );
    cvsUpdateAction().setEnabled( hasDocument || hasExistingSelection );
    cvsCommitAction().setEnabled( hasDocument || hasExistingSelection );

}

//______________________________________________________
void FileSelectionFrame::_updateNavigationActions( void )
{
    Debug::Throw( "FileSelectionFrame::_updateNavigationActions.\n" );
    previousDirectoryAction().setEnabled( editor_->hasPrevious() );
    nextDirectoryAction().setEnabled( editor_->hasNext() );
    parentDirectoryAction().setEnabled( editor_->hasParent() );
    return;
}

//______________________________________________________
void FileSelectionFrame::_showToolTip( const QModelIndex& index )
{
    if( !index.isValid() ) toolTipWidget_->hide();
    else {

        // fileInfo
        const FileInfo fileInfo( model_.get( index ) );

        // icon
        QIcon icon;
        QVariant iconVariant( model_.data( index, Qt::DecorationRole ) );
        if( iconVariant.canConvert( QVariant::Icon ) ) icon = iconVariant.value<QIcon>();

        // rect
        QRect rect( list_->visualRect( index ) );
        rect.translate( list_->viewport()->mapToGlobal( QPoint( 0, 0 ) ) );

        // assign to tooltip widget
        toolTipWidget_->setFileInfo( fileInfo, icon );
        toolTipWidget_->adjustPosition( rect );
        toolTipWidget_->showDelayed();

    }

}

//______________________________________________________
void FileSelectionFrame::_reload( bool clearFirst, bool checkCvs )
{

    Debug::Throw( "FileSelectionFrame::_reload.\n" );
    if( clearFirst ) clear();
    if( path().isEmpty() ) return;

    // loop over directory contents
    unsigned int flags = File::None;
    if( hiddenFilesAction().isChecked() ) flags |= File::ShowHiddenFiles;
    File::List files( path().listFiles( flags ) );

    FileInfo::List newFiles;
    foreach( const File& file, files )
    {

        // propagate file properties to FileInfo
        FileInfo fileInfo( file );
        if( file.isDirectory() ) fileInfo.setIsFolder();
        else fileInfo.setIsDocument();

        if( file.isLink() ) fileInfo.setIsLink();
        if( file.isBrokenLink() ) fileInfo.setIsBrokenLink();
        if( file.isHidden() ) fileInfo.setIsHidden();
        fileInfo.update();

        // add to model
        newFiles << fileInfo;

    }

    // get files that have been removed
    FileInfo::List deletedFiles;
    foreach( const FileInfo& fileInfo, model_.get() )
    {
        if( std::find_if( newFiles.begin(), newFiles.end(), FileInfo::SameFileFTor( fileInfo.file() ) ) == newFiles.end() )
        { deletedFiles << fileInfo; }
    }

    // update model and list
    model_.add( newFiles );
    model_.remove( deletedFiles );
    if( AnimatedTreeView* treeView = qobject_cast<AnimatedTreeView*>( list_ ) )
    { treeView->resizeColumnToContents( FileInfoModel::FILE ); }

    // try select local file if any
    _selectFile();

    // run cvs update command
    if( checkCvs )
    {
        if( interface().isCvsPath( path() ) )
        {

            // try read cache
            _showStatusFromLocalCache();

            // reload from interface (delayed to avoid running too many CVS commands)
            timer_.start( 500, this );

        } else emit messageAvailable( "[not a CVS path]\n", CvsInterface::LOW );

    }

}

//______________________________________________________
void FileSelectionFrame::timerEvent( QTimerEvent* event )
{

    Debug::Throw() << "FileSelectionFrame::timerEvent - path: " << path() << endl;

    if( event->timerId() == timer_.timerId() )
    {
        timer_.stop();

        // update cvs root and repository
        interface().updateRepositoryPath( path() );

        // update using interface
        interface().abort(CvsCommand::STATUS|CvsCommand::FAKE_UPDATE);
        interface().logStatus( path() );
        interface().fakeUpdate( path() );

    } else return QObject::timerEvent( event );

};

//______________________________________________________
void FileSelectionFrame::_selectFile( void )
{
    Debug::Throw( "FileSelectionFrame::_selectFile.\n" );
    if( file_.isEmpty() )  return;

    const FileInfoModel::List& indexes( model_.get() );
    FileInfoModel::List::const_iterator iter( std::find_if( indexes.begin(), indexes.end(), FileInfo::SameFileFTor( file_ ) ) );
    if( iter != indexes.end() )
    {
        QModelIndex index( model_.index( *iter ) );
        list_->selectionModel()->select( index, QItemSelectionModel::ClearAndSelect );
        list_->scrollTo( index );
    }

}

//______________________________________________________
void FileSelectionFrame::_edit( void )
{
    Debug::Throw( "FileSelectionFrame::_edit.\n" );
    FileInfoModel::List selection( this->selection() );
    foreach( const FileInfo& fileInfo, selection )
    { if( fileInfo.type() & FileInfo::Document ) _edit( fileInfo.file() ); }
}

//______________________________________________________
void FileSelectionFrame::_remove( void )
{

    Debug::Throw( "FileSelectionFrame::_remove.\n" );

    // retrieve selected items; make sure they do not include the navigator
    FileInfoModel::List selection( this->selection() );
    selection.erase( std::remove_if( selection.begin(), selection.end(), FileInfo::SameTypeFTor(  FileInfo::Navigator ) ), selection.end() );

    // open confirmation dialog
    RemoveFilesDialog dialog( this, selection );
    dialog.setWindowTitle( "Delete Files - Cvswatch" );
    if( !dialog.exec() ) return;

    // retrieve selected items and remove corresponding files
    foreach( const FileInfo& fileInfo, selection )
    {
        const File file( fileInfo.file().addPath( path() ) );
        if( file.exists() ) file.removeRecursive();
    }

    // update (need full reload here to update cache and because this may not be a CVS directory)
    _reload();

}

//______________________________________________________
void FileSelectionFrame::_previousDirectory( void )
{
    Debug::Throw( "FileSelectionFrame::_previousDirectory.\n" );
    if( !editor_->hasPrevious() ) return;
    editor_->selectPrevious();
    setFile( editor_->path() );
    emit fileSelected( path() );
}

//______________________________________________________
void FileSelectionFrame::_nextDirectory( void )
{
    Debug::Throw( "FileSelectionFrame::_nextDirectory.\n" );
    if( !editor_->hasNext() ) return;
    editor_->selectNext();
    setFile( editor_->path() );
    emit fileSelected( path() );
}

//______________________________________________________
void FileSelectionFrame::_parentDirectory( void )
{

    Debug::Throw( "FileSelectionFrame::_parentDirectory.\n" );
    editor_->selectParent();
    setFile( editor_->path() );
    emit fileSelected( path() );

}

//______________________________________________________
void FileSelectionFrame::_homeDirectory( void )
{
    Debug::Throw( "FileSelectionFrame::_homeDirectory.\n" );
    setFile( Util::home() );
    emit fileSelected( path() );
}

//______________________________________________________
void FileSelectionFrame::_showStatus( File path, FileInfo::List files )
{
    Debug::Throw() << "FileSelectionFrame::_showStatus - path: " << path << endl;

    // update cache
    // this is an incremental update rather than a full replacement
    // because _showStatus can be called sequentially rather than by blocks.
    foreach( const FileInfo& fileInfo, files )
    {

        Debug::Throw() << "FileSelectionFrame::_showStatus - found " << fileInfo.file() << endl;

        // don't add the "locally removed" nor "remotely added" files to the cache because at the point where
        // cvs does not know about them any-more (after cvs commit) there is no way to remove
        // them from the cache.
        if( !( fileInfo.status() & (FileInfo::LocallyRemoved|FileInfo::RemotelyAdded) ) && ( fileInfo.type() & FileInfo::Document ) )
        { localCache_.insert( fileInfo ); }

    }

    // check if path has changed
    // if yes, update, then reset local cache and return
    if( !FileSelectionFrame::path().isEqual( path ) )
    {

        Debug::Throw( "FileSelectionFrame::_showStatus - invalid file.\n" );
        cache_.merge( localCache_ );
        localCache_ = cache_.findAll( FileSelectionFrame::path() );
        return;

    } else Debug::Throw() << "FileSelectionFrame::_showStatus - matching path: " << path << endl;

    // loop over existing items
    FileInfoModel::List currentFiles( model_.get() );
    FileInfoModel::List newFiles;
    FileInfoModel::List removedFiles;
    foreach( const FileInfo& fileInfo, currentFiles )
    {

        Debug::Throw() << "FileSelectionFrame::_showStatus - processing " << fileInfo.file() << endl;

        // everything exept documents is skipped
        if( !(fileInfo.type() & FileInfo::Document ) ) continue;

        FileInfo::List::iterator found( std::find_if( files.begin(), files.end(), FileInfo::SameFileFTor( fileInfo.file() ) ) );
        if( found != files.end() )
        {

            Debug::Throw() << "FileSelectionFrame::_showStatus - fullname: " << found->file() << endl;
            if( found->file().exists() )
            {

                // if file exists, set fileInfo time stamp to its last modification time
                // this avoids some glitches when clocks are not perfectly synchronized
                // between remote repository and local time
                found->setCreation( fileInfo.lastModified() );
                localCache_.insert( *found  );

                // this could be avoided by copying local information from existing info.
                found->updateFrom( fileInfo );
                newFiles << *found;

            } else if( found->status() == FileInfo::RemotelyModified ) {

                // changed from modified to added since the file was not found locally
                // update cache accordingly
                found->setStatus( FileInfo::RemotelyAdded );
                localCache_.insert( *found );
                found->update();
                newFiles << *found;

            } else if(  found->status() == FileInfo::UpToDate ) {

                // up-to-date files that cannot be found locally correspond to successfully files succesfully removed from the list
                // one can safely remove them from the list, and from the global cache
                localCache_.remove( *found );
                cache_.remove( *found );
                removedFiles.push_back( *found );

            } else {

                found->update();
                newFiles << *found;

            }

            // file has been addded. Remove from list
            files.erase( found );

        };

    }

    // add remaining items (that corresponds to locally removed files)
    foreach( FileInfo fileInfo, files )
    {

        // check path: first add the current path, retrieve the full path
        // (because the iterator may have a local path included)
        // and compare
        Debug::Throw() << "FileSelectionFrame::showStatus:"
            << " item path: " << fileInfo.file().path()
            << " this path: " << FileSelectionFrame::path()
            << endl;
        if( !fileInfo.file().path().isEqual( FileSelectionFrame::path() ) )
        { continue; }

        // everything exept from documents is skipped
        if( !(fileInfo.type() & FileInfo::Document ) ) continue;

        if( fileInfo.status() == FileInfo::RemotelyModified && !fileInfo.file().exists() )
        {

            // modify status from remotely modified to remotely added
            // if file does not exists
            fileInfo.setStatus( FileInfo::RemotelyAdded );
            localCache_.insert( fileInfo );

        }

        // create new item
        if( fileInfo.status() != FileInfo::UpToDate )
        {
            fileInfo.update();
            newFiles << fileInfo;
        }

    }

    // update model
    model_.remove( removedFiles );
    model_.add( newFiles );

    if( !( removedFiles.empty() && newFiles.empty() ) )
    {

        if( AnimatedTreeView* treeView = qobject_cast<AnimatedTreeView*>( list_ ) )
        { treeView->resizeColumnToContents( FileInfoModel::FILE ); }
        _selectFile();

    }

}

//______________________________________________________
void FileSelectionFrame::_cvsCheckOut( void )
{
    Debug::Throw( "FileSelectionFrame::_cvsCheckOut.\n" );
    CvsCheckOutDialog dialog( this );
    dialog.setWindowTitle( "Checkout Files - Cvswatch" );

    QString root( interface().root() );
    if( root.isEmpty() ) root = XmlOptions::get().raw( "DEFAULT_CVSROOT" );
    dialog.setCvsRoot( root );
    if( dialog.exec() == QDialog::Rejected ) return;

    // run command
    interface().checkOut( path(), dialog.cvsRoot(), dialog.modules(), File( dialog.useLocalName() ? dialog.localName():"" ) );

    // update CVSRoot base option from dialog
    XmlOptions::get().set( "DEFAULT_CVSROOT", dialog.cvsRoot() );

    return;

}

//______________________________________________________
void FileSelectionFrame::_cvsUpdate( void )
{

    Debug::Throw( "FileSelectionFrame::_cvsUpdate.\n" );
    FileInfoModel::List selection( this->selection() );
    interface().update( path(), selection );
    interface().logStatus( path(), selection );

}

//______________________________________________________
void FileSelectionFrame::_cvsUpdateDirectory( void )
{

    Debug::Throw( "FileSelectionFrame::_cvsUpdateDirectory.\n" );
    interface().update( path() );
    interface().logStatus( path() );

}

//______________________________________________________
void FileSelectionFrame::_cvsAdd( void )
{

    Debug::Throw( "FileSelectionFrame::_cvsAdd.\n" );

    // retrieve selected items; make sure they do not include the navigator or links
    FileInfoModel::List selection( this->selection() );
    selection.erase( std::remove_if( selection.begin(), selection.end(), FileInfo::SameTypeFTor( FileInfo::Navigator ) ), selection.end() );
    selection.erase( std::remove_if( selection.begin(), selection.end(), FileInfo::SameTypeFTor( FileInfo::Link ) ), selection.end() );

    // create dialog
    CvsAddDialog dialog( this, selection );
    dialog.setWindowTitle( "Add Files - Cvswatch" );
    if( dialog.exec() == QDialog::Rejected  || ( selection = dialog.selectedFiles() ).empty() ) return;

    // run cvs
    interface().add( path(), selection );
    interface().logStatus( path(), selection );
    return;

}

//______________________________________________________
void FileSelectionFrame::_cvsRemove( void )
{

    Debug::Throw( "FileSelectionFrame::_cvsRemove.\n" );

    // retrieve selected items; make sure they only include documents
    FileInfoModel::List selection( this->selection() );
    selection.erase( std::remove_if( selection.begin(), selection.end(), FileInfo::DifferentTypeFTor( FileInfo::Document ) ), selection.end() );

    // create dialog
    CvsRemoveDialog dialog( this, selection );
    dialog.setWindowTitle( "Remove Files - Cvswatch" );
    dialog.setMode( CvsRemoveDialog::REMOVE_FILES );
    if( dialog.exec() == QDialog::Rejected  || ( selection = dialog.selectedFiles() ).empty() ) return;

    // retrieve dialog mode
    CvsRemoveDialog::Mode mode( dialog.mode() );

    foreach( const FileInfo& fileInfo, selection )
    {

        // retrieve file
        File oldFile( fileInfo.file().addPath( path() ) );

        // check file exists
        if( !oldFile.exists() ) continue;

        // remove file if required
        if( mode == CvsRemoveDialog::REMOVE_FILES ) oldFile.remove();
        else oldFile.rename( oldFile.truncatedName() + "_removed." + oldFile.extension() );

    }

    // remove files from cvs
    interface().remove( path(), selection );
    interface().logStatus( path(), selection );

    // update (need full reload here to update cache and because this may not be a CVS directory)
    _reload();

}

//______________________________________________________
void FileSelectionFrame::_cvsCommit( void )
{
    Debug::Throw( "FileSelectionFrame::_cvsCommit.\n" );

    // retrieve selected items; remove navigator and linked documents
    FileInfoModel::List selection(this->selection() );
    selection.erase( std::remove_if( selection.begin(), selection.end(), FileInfo::SameTypeFTor( FileInfo::Navigator ) ), selection.end() );
    selection.erase( std::remove_if( selection.begin(), selection.end(), FileInfo::SameTypeFTor( FileInfo::Document | FileInfo::Link ) ), selection.end() );

    // create dialog
    CvsCommitDialog dialog( this, selection );
    dialog.setWindowTitle( "Commit Files - Cvswatch" );
    if( dialog.exec() == QDialog::Rejected || (selection = dialog.selectedFiles()).empty() ) return;

    // retrieve comments and put in a temporary file
    // the file has a unique name and is removed when the commit is performed
    File tmp;
    QString comments( dialog.editor().toPlainText().toAscii() );
    if( !comments.isEmpty() )
    {
        QString buffer;
        QTextStream( &buffer ) << "_cvs_commit_" << Util::user() << "_" << TimeStamp::now().unixTime() << "_" << Util::pid() << ".tmp";
        tmp = File( buffer ).addPath( Util::tmp() );
        QFile in( tmp );
        if( in.open( QIODevice::WriteOnly ) )
        {
            in.write( dialog.editor().toPlainText().toAscii() );
            in.close();
        }
    }

    interface().commit( path(), tmp, selection );
    interface().logStatus( path(), selection );

}


//______________________________________________________
void FileSelectionFrame::_cvsCommitDirectory( void )
{
    Debug::Throw( "FileSelectionFrame::_cvsCommitDirectory.\n" );

    // create dialog
    FileInfo fileInfo( path(), FileInfo::Folder );
    fileInfo.update();
    CvsCommitDialog dialog( this, fileInfo );
    dialog.setWindowTitle( "Commit Directory - Cvswatch" );
    if( dialog.exec() == QDialog::Rejected ) return;

    // retrieve selected items
    if( dialog.selectedFiles().empty() ) return;

    // retrieve comments and put in a temporary file
    // the file has a unique name and is removed when the commit is performed
    File tmp;
    QString comments( dialog.editor().toPlainText().toAscii() );
    if( !comments.isEmpty() )
    {
        QString buffer;
        QTextStream( &buffer ) << "_cvs_commit_" << Util::user() << "_" << TimeStamp::now().unixTime() << "_" << Util::pid() << ".tmp";
        tmp = File( buffer ).addPath( Util::tmp() );
        QFile in( tmp );
        if( in.open( QIODevice::WriteOnly ) )
        {
            in.write( dialog.editor().toPlainText().toAscii() );
            in.close();
        }
    }

    interface().commit( path(), tmp );
    interface().logStatus( path() );

}

//________________________________________
void FileSelectionFrame::_storeSelection( void )
{ model_.setSelectedIndexes( list_->selectionModel()->selectedRows() ); }


//________________________________________
void FileSelectionFrame::_restoreSelection( void )
{

    const QModelIndexList selection( model_.selectedIndexes() );
    list_->selectionModel()->clearSelection();

    foreach( const QModelIndex& index, selection )
    { list_->selectionModel()->select( index, QItemSelectionModel::Select|QItemSelectionModel::Rows ); }

    return;

}

//____________________________________________________________________________________________
void FileSelectionFrame::_toggleShowNavigator( bool value )
{

    Debug::Throw() << "FileSelectionFrame::_toggleShowNavigator - " << value << endl;
    showNavigator_ = value;
    _setShowNavigator( showNavigator_ && viewMode_ == LIST );

}

//______________________________________________________
void FileSelectionFrame::_setShowNavigator( bool value )
{
    FileInfo navigator( File( ".." ), FileInfo::Navigator );
    navigator.update();
    if( value ) model_.add( navigator );
    else model_.remove( navigator );
}

//_____________________________________________
void FileSelectionFrame::_installActions( void )
{

    Debug::Throw( "FileSelectionFrame::_installActions.\n" );

    // hidden files
    addAction( hiddenFilesAction_ = new QAction( "Show Hidden Files", this ) );
    hiddenFilesAction_->setCheckable( true );
    connect( hiddenFilesAction_, SIGNAL( toggled( bool ) ), SLOT( _toggleShowHiddenFiles( bool ) ) );

    // edit
    addAction( editAction_ = new QAction( IconEngine::get( ICONS::EDIT ), "Edit", this ) );
    connect( editAction_, SIGNAL( triggered() ), SLOT( _edit() ) );

    // remove
    addAction( removeAction_ = new QAction( IconEngine::get( ICONS::DELETE ), "Delete", this ) );
    connect( removeAction_, SIGNAL( triggered() ), SLOT( _remove() ) );
    removeAction_->setShortcut( QKeySequence::Delete );
    removeAction_->setToolTip( "Remove selected files locally" );

    // previous directory
    addAction( previousDirectoryAction_ = new QAction( IconEngine::get( ICONS::PREVIOUS_DIRECTORY ), "Previous", this ) );
    connect( previousDirectoryAction_, SIGNAL( triggered() ), SLOT( _previousDirectory() ) );
    previousDirectoryAction_->setToolTip( "Change path to previous directory (from history)" );

    // next directory (from history)
    addAction( nextDirectoryAction_ = new QAction( IconEngine::get( ICONS::NEXT_DIRECTORY ), "Next", this ) );
    connect( nextDirectoryAction_, SIGNAL( triggered() ), SLOT( _nextDirectory() ) );
    nextDirectoryAction_->setToolTip( "Change path to next directory (from history)" );

    // parent directory in tree
    addAction( parentDirectoryAction_ = new QAction( IconEngine::get( ICONS::PARENT_DIRECTORY ), "Parent Directory", this ) );
    connect( parentDirectoryAction_, SIGNAL( triggered() ), SLOT( _parentDirectory() ) );
    parentDirectoryAction_->setToolTip( "Change path to parent directory" );

    // home directory
    addAction( homeDirectoryAction_ = new QAction( IconEngine::get( ICONS::HOME_DIRECTORY ), "Home", this ) );
    connect( homeDirectoryAction_, SIGNAL( triggered() ), SLOT( _homeDirectory() ) );
    homeDirectoryAction_->setToolTip( "Change path to user home" );

    // cvs checkout
    addAction( cvsCheckoutAction_ = new QAction( IconEngine::get( ICONS::REMOTELY_ADDED ), "Checkout", this ) );
    connect( cvsCheckoutAction_, SIGNAL( triggered() ), SLOT( _cvsCheckOut() ) );

    // cvs update
    addAction( cvsUpdateAction_ = new QAction( IconEngine::get( ICONS::REMOTELY_MODIFIED ), "Update", this ) );
    connect( cvsUpdateAction_, SIGNAL( triggered() ), SLOT( _cvsUpdate() ) );

    // cvs update
    addAction( cvsAddAction_ = new QAction( IconEngine::get( ICONS::LOCALLY_ADDED ), "Add", this ) );
    connect( cvsAddAction_, SIGNAL( triggered() ), SLOT( _cvsAdd() ) );

    // cvs update
    addAction( cvsRemoveAction_ = new QAction( IconEngine::get( ICONS::LOCALLY_REMOVED ), "Remove", this ) );
    connect( cvsRemoveAction_, SIGNAL( triggered() ), SLOT( _cvsRemove() ) );

    // cvs commit
    addAction( cvsCommitAction_ = new QAction( IconEngine::get( ICONS::LOCALLY_MODIFIED ), "Commit", this ) );
    connect( cvsCommitAction_, SIGNAL( triggered() ), SLOT( _cvsCommit() ) );

    // cvs update directory
    addAction( cvsUpdateDirectoryAction_ = new QAction( IconEngine::get( ICONS::REMOTELY_MODIFIED ), "Update Directory", this ) );
    connect( cvsUpdateDirectoryAction_, SIGNAL( triggered() ), SLOT( _cvsUpdateDirectory() ) );

    // cvs commit directory
    addAction( cvsCommitDirectoryAction_ = new QAction( IconEngine::get( ICONS::LOCALLY_MODIFIED ), "Commit Directory", this ) );
    connect( cvsCommitDirectoryAction_, SIGNAL( triggered() ), SLOT( _cvsCommitDirectory() ) );

}

//______________________________________________________
void FileSelectionFrame::_showStatusFromLocalCache( void )
{
    Debug::Throw( "FileSelectionFrame::_showStatusFromLocalCache.\n" );

    // make a copy of local cache (because items found in the list()
    // are removed on fly
    FileInfoCache files( localCache_ );

    // loop over items
    FileInfoModel::List currentFiles( model_.get() );
    FileInfoModel::List newFiles;
    foreach( FileInfo fileInfo, currentFiles )
    {

        FileInfoCache::iterator found( std::find_if( files.begin(), files.end(), FileInfo::SameFileFTor( fileInfo.file() ) ) );
        if( found == files.end() )
        {

            if( fileInfo.type() == FileInfo::Document )
            {
                fileInfo.setStatus( FileInfo::Unknown );
                newFiles << fileInfo;
            }

        } else if(
            fileInfo.file().addPath( path() ).lastModified() > found->created() &&
            (fileInfo.type() & FileInfo::Document) &&
            (fileInfo.status() == FileInfo::UpToDate) )
        {

            fileInfo.setStatus( FileInfo::LocallyModified );
            newFiles << fileInfo;

            localCache_.remove( *found );
            cache_.remove( *found );
            files.erase( found );

        } else {

            fileInfo.setStatus( found->status() );
            fileInfo.update();
            newFiles << fileInfo;
            files.erase( found );

        }

    }

    // update model
    model_.add( newFiles );

}

//___________________________________________________________________________________
void FileSelectionFrame::_edit( File file )
{
    Debug::Throw() << "FileSelectionFrame::_edit - file: " << file << endl;
    ( Command( XmlOptions::get().raw( "EDITOR" ) ) << file.addPath( path() ).expand() ).run();
}
